from datetime import datetime
import logging

from .common import KNOWN_SITES as common_known, get_publisher, DataFetcher
from ..classes import CslItem, CslType, CslName

KNOWN_SITES = common_known.copy()
KNOWN_SITES.update({"pypi": "The Python Package Index"})

logger = logging.getLogger(__name__)


class PypiDataFetcher(DataFetcher):
    base_url = "https://www.pypi.org/pypi"

    def get_authors(self, info):
        author_str = info.get("author")
        maintainer_str = info.get("maintainer")

        authors = []

        if author_str:
            authors.append(CslName(literal=author_str))
        elif maintainer_str:
            authors.append(CslName(literal=maintainer_str))

        return authors

    async def get(self, package, version=None, date_accessed=None) -> CslItem:
        url = self.base_url + "/" + package
        if version:
            url += "/" + version
        url += "/" + "json"

        logger.debug("Fetching information from %s", url)

        response = await self.client.get(url)
        response.raise_for_status()
        data = response.json()
        logger.debug("Successfully parsed data")
        info = data["info"]

        release = data["releases"][info["version"]][0]
        dt = datetime.fromisoformat(release["upload_time"])

        first_upload = dt
        for release in data["releases"].values():
            for upload in release:
                first_upload = min(
                    first_upload, datetime.fromisoformat(upload["upload_time"])
                )

        item_url = info.get("home_page") or info["project_url"]
        publisher = get_publisher(item_url, KNOWN_SITES)

        return CslItem(
            type=CslType.WEBPAGE,
            id=package,
            author=self.get_authors(info),
            URL=item_url,
            abstract=info["summary"] or None,
            version=info["version"],
            issued=dt,
            # event_date=dt,
            # container=dt,
            # submitted=dt,
            original_date=first_upload,
            accessed=date_accessed,
            categories=(
                ["software", "python", "libraries", "pypi"] + info["classifiers"]
            ),
            publisher=publisher,
            title=package,
        )



from datetime import datetime
import logging

from .common import DataFetcher, KNOWN_SITES as common_known, get_publisher
from ..classes import CslItem, CslType, CslName


KNOWN_SITES = common_known.copy()
KNOWN_SITES.update({"crates.io": "Crates.io", "docs.rs": "Docs.rs"})


logger = logging.getLogger(__name__)


class CratesDataFetcher(DataFetcher):
    base_url = "https://www.crates.io"

    async def get_date_author(self, version_dict):
        issued = datetime.fromisoformat(version_dict["created_at"])

        authors_url = self.base_url + version_dict["links"]["authors"]
        author_response = await self.client.get(authors_url)
        author_response.raise_for_status()
        names = author_response.json()["meta"]["names"]
        return issued, [CslName(literal=name) for name in names]

    async def get(self, package, version=None, date_accessed=None) -> CslItem:
        api_url = self.base_url + "/api/v1/crates/" + package

        logger.debug("Fetching information from %s", api_url)

        response = await self.client.get(api_url)
        response.raise_for_status()
        data = response.json()
        crate_data = data["crate"]

        categories = ["software", "rust", "libraries", "crates"]
        categories.extend(crate_data.get("keywords", []))
        categories.extend(crate_data.get("categories", []))

        item_url = crate_data.get("homepage")

        if item_url is None:
            item_url = "/".join([self.base_url, "crates", package])
            if version:
                item_url += "/" + version

        if not version:
            version = crate_data["max_version"]

        original_authors = None
        issued = None
        authors = None

        for idx, v_dict in enumerate(data["versions"]):
            if idx == len(data["versions"]) - 1:
                _, original_authors = await self.get_date_author(v_dict)

            if v_dict["num"] == version:
                issued, authors = await self.get_date_author(v_dict)

        return CslItem(
            type=CslType.WEBPAGE,
            id=package,
            abstract=data.get("description", ""),
            author=authors,
            URL=item_url,
            version=version,
            issued=issued,
            original_author=original_authors,
            original_date=datetime.fromisoformat(crate_data["created_at"]),
            accessed=date_accessed,
            categories=categories,
            publisher=get_publisher(item_url, KNOWN_SITES),
            title=crate_data.get("name", package),
        )



import argparse
import json
import sys
import logging
from typing import Dict, Optional
import re
from contextlib import contextmanager
from pip._internal.operations.freeze import freeze as pip_freeze
import textwrap
import asyncio
import datetime as dt
import os

import httpx

from citepy import __version__
from citepy.repos import KNOWN_FETCHERS

logger = logging.getLogger(__name__)

DATE_ACCESSED_VAR = "CITEPY_DATE_ACCESSED"
DEFAULT_DATE_STR = os.environ.get(DATE_ACCESSED_VAR, dt.date.today().isoformat())
DEFAULT_JOBS = 5


def get_pypi_versions() -> Dict[str, Optional[str]]:
    """Get versions from ``pip freeze``"""
    d = dict()
    for line in pip_freeze():
        try:
            k, v = line.strip().split("==")
        except ValueError:
            continue
        d[k] = v
    return d


def setup_logging(verbosity):
    verbosity = verbosity or 0
    levels = [logging.WARNING, logging.INFO, logging.DEBUG, logging.NOTSET]
    logging.basicConfig(level=levels[min(verbosity, len(levels) - 1)])

    levels.insert(0, logging.CRITICAL)
    loud_level = levels[min(verbosity, len(levels) - 1)]
    for name in ["pip", "urllib3", "websockets"]:
        logging.getLogger(name).setLevel(loud_level)


def msg_template(e):
    return "Could not get citation for package '%s'\n" + textwrap.indent(
        str(e), " " * 4
    )


name_re = re.compile(
    r"^(?P<name>(\w[\w\d\._-]*))\s*((?P<rel>[=><!~^]{1,2})\s*(?P<ver>[\d\.\*\w-]+))?$"
)


def split_package_versions(packages):
    for s in packages:
        s = s.strip()
        if s == "-":
            yield from split_package_versions(sys.stdin.readlines())
            continue

        m = name_re.match(s)
        if m is None:
            logger.warning("Could not parse '%s'; skipping", s)
            continue

        g = m.groupdict()
        name = g["name"]
        rel = g["rel"]
        ver = g["ver"]

        if rel:
            if set(rel) == {"="} and len(rel) <= 2:
                yield name, ver
                continue
            else:
                logger.warning(
                    f"Unsupported package-version relationship '{rel}'; "
                    "ignoring version"
                )
        yield name, None


async def get_info(
    package_versions: Dict[str, Optional[str]], repo: str, date: dt.date = None
):
    fetcher_cls = KNOWN_FETCHERS[repo]
    # sem = asyncio.Semaphore(jobs)
    futs = []
    async with httpx.AsyncClient() as c:
        fetcher = fetcher_cls(c)
        for k, v in package_versions.items():
            # async with sem:
            futs.append(fetcher.get(k, v, date))
        results = await asyncio.gather(*futs)
    return results


@contextmanager
def outfile(obj):
    if not obj or obj == "-":
        yield sys.stdout
    else:
        with open(obj, "w") as f:
            yield f


@contextmanager
def infile(obj):
    if not obj or obj == "-":
        yield sys.stdin
    else:
        with open(obj) as f:
            yield f


def parse_date(s: str) -> dt.date:
    datetime = dt.datetime.strptime(s, "%Y-%m-%d")
    return datetime.date()


def read_packages(args):
    if not args:
        return
    has_stdin = False
    for arg in args:
        if arg == "-":
            if has_stdin:
                continue
            else:
                has_stdin = True
        with infile(arg) as f:
            for line in f:
                stripped = line.strip()
                if not stripped or line.startswith("#"):
                    continue
                yield stripped


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "package",
        nargs="*",
        help=(
            "names of packages you want to cite, "
            "optionally with (full) version string. "
            "e.g. numpy==1.16.3 beautifulsoup4==4.7.1 . "
            "Note that version strings are handled differently "
            "by different repositories, and may be ignored. "
            "In particular, any non-exact version constraint is ignored."
        ),
    )
    parser.add_argument(
        "--all-python",
        "-a",
        action="store_true",
        help=(
            "if set, will get information for all python packages "
            "accessible to `pip freeze`"
        ),
    )
    parser.add_argument(
        "--repo",
        "-r",
        default="pypi",
        choices=sorted(KNOWN_FETCHERS),
        help="which package repository to use (default pypi)",
    )
    parser.add_argument(
        "--infile",
        "-i",
        action="append",
        help=(
            "path to read input packages from as newline-separated items "
            "(can be given multiple times; - reads from stdin)"
        ),
    )
    parser.add_argument(
        "--outfile",
        "-o",
        help="path to write output to (default or - writes to stdout)",
    )
    # parser.add_argument(
    #     "--format",
    #     "-f",
    #     default="csl-json",
    #     choices=("csl-json",),
    #     help="format to write out (default 'csl-json')",
    # )
    # parser.add_argument(
    #     "--jobs",
    #     "-j",
    #     type=int,
    #     default=DEFAULT_JOBS,
    #     help=f"Number of concurrent requests to make (default {DEFAULT_JOBS})",
    # )
    parser.add_argument(
        "--verbose",
        "-v",
        action="count",
        help="Increase verbosity of logging (can be repeated). "
        "One for DEBUG, two for NOTSET, three includes all library logging.",
    )
    parser.add_argument(
        "--date-accessed",
        "-d",
        type=parse_date,
        default=parse_date(DEFAULT_DATE_STR),
        help=(
            "Manually set access date, in format 'YYYY-MM-DD'. "
            f"Falls back to {DATE_ACCESSED_VAR} environment variable, "
            "then today's date."
        ),
    )
    parser.add_argument(
        "--version", action="store_true", help="print version information and exit"
    )

    parsed = parser.parse_args()

    setup_logging(parsed.verbose)

    if parsed.version:
        print(__version__)
        sys.exit(0)

    parsed.package.extend(read_packages(parsed.infile))

    if parsed.repo == "pypi":
        versions = get_pypi_versions()
        if parsed.package:
            package_versions = {
                p: v or versions.get(p)
                for p, v in split_package_versions(parsed.package)
            }
        else:
            package_versions = versions
    else:
        package_versions = dict(split_package_versions(parsed.package))

    csl_items = asyncio.run(
        get_info(package_versions, parsed.repo, parsed.date_accessed)
    )  # , parsed.jobs))
    s = json.dumps([item.to_jso() for item in csl_items], indent=2, sort_keys=True)

    # if parsed.format.lower() == "json":
    #     s = json.dumps(
    #         [item.to_jso() for item in csl_items],
    #         indent=2,
    #         sort_keys=True
    #     )
    # else:
    #     raise ValueError(f"Unrecognised output format '{parsed.format}'")

    with outfile(parsed.outfile) as f:
        print(s, file=f)

    sys.exit(0)


if __name__ == "__main__":
    main()



from typing import Dict, Type

from .common import DataFetcher
from .pypi import PypiDataFetcher
from .crates import CratesDataFetcher


KNOWN_FETCHERS: Dict[str, Type[DataFetcher]] = {
    "pypi": PypiDataFetcher,
    "crates": CratesDataFetcher,
}

__all__ = ["PypiDataFetcher", "CratesDataFetcher", "KNOWN_FETCHERS", "DataFetcher"]



#!/usr/bin/env python3
from pathlib import Path

from setuptools import setup

here = Path(__file__).absolute().parent

with open(here / "README.md") as f:
    long_description = f.read()

setup(
    name="citepy",
    packages=["citepy", "citepy.repos"],
    url="https://www.github.com/clbarnes/citepy",
    license="MIT",
    author="Chris L. Barnes",
    author_email="cbarnes@mrc-lmb.cam.ac.uk",
    description="Automatically create citations for packages",
    install_requires=["jsonschema", "httpx"],
    package_data={"citepy": ["csl-data.json"]},
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires=">=3.7,<4",
    entry_points={
        "console_scripts": ["citepy = citepy.cli:main", "cite = citepy.cli:main"]
    },
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
)
